from pydantic import Field

from ..types import Tier
from .base_model import BaseModel


class Item(BaseModel):
    """Parent class for Battle and Held items.

    See [here](https://unite-db.com/items) for a list of all items on
    the unite-db website.
    """

    name: str = Field(..., description="Name of this item.")
    display_name: str = Field(..., description="Display name of this item.")
    tier: Tier = Field(..., description="What tier unite-db consider this item to be.")
    description: str = Field(..., description="Description of this item.")



from .mock_unite_db import MockUniteDb

aeos_cookie = MockUniteDb().held_items[0]


def test_effect_at():
    assert aeos_cookie.effect.effect_at(1) == 100
    assert aeos_cookie.effect.effect_at(10) == 150
    assert aeos_cookie.effect.effect_at(20) == 200

    assert aeos_cookie.scaling_effects[0].effect_at(1) == 8
    assert aeos_cookie.scaling_effects[0].effect_at(15) == 120
    assert aeos_cookie.scaling_effects[0].effect_at(30) == 240


def test_description_at():
    assert aeos_cookie.effect.description_at(1) == "100 Max HP"
    assert aeos_cookie.scaling_effects[0].description_at(1) == "HP +8"



from typing import Any, Optional

from pydantic import Field

from ..types import DamageType, Difficulty, Range, Role, Tier
from ..utils import _flatten
from .base_model import BaseModel
from .stats import Stats

POKEMON_MIN_LEVEL = 1
POKEMON_MAX_LEVEL = 15


class Evolution(BaseModel):
    name: str
    level: int


class Stars(BaseModel):
    combat: int
    resistance: int
    mobility: int
    scoring: int
    assistance: int
    total: int


class Pokemon(BaseModel):
    """Class representing a Pokemon.

    See all Pokemon [here](https://unite-db.com) on the unite-db website.

    Get all pokemon from the API using
    [`UniteDb.pokemon`][py_unite_db.UniteDb].

    Example:
        >>> from py_unite_db import UniteDb
        >>> unite_db = UniteDb()

        >>> unite_db.pokemon[0].name
        'Absol'
        >>> unite_db.pokemon[0].tier
        'C'
    """

    name: str = Field(..., description="Name of the Pokemon.")
    notes: str = Field("", description="Things to note about the Pokemon.")
    tier: Tier = Field(
        ...,
        description="""What tier unite-db consider this Pokemon to be.
    From S being the top tier, to D being the lowest.
    T is for Pokemon coming soon.""",
    )
    damage_type: DamageType = Field(..., description="Physical or special attacker.")
    stars: Optional[Stars] = Field(
        ..., description="How many stars the pokemon has for each stat in the game."
    )
    stats: list[Stats] = Field([], description="This Pokemon's stats.")
    range: Range = Field(..., description="Ranged or Melee attacker.")
    difficulty: Difficulty = Field(
        ..., description="The in-game difficulty rating of the Pokemon."
    )
    role: Role = Field(..., description="The role of the Pokemon.")
    soon: bool = Field(..., description="Whether the Pokemon is coming soon or not.")
    evolution: list[Evolution] = Field(
        [], description="List of the Pokemon's evolutions."
    )

    def stats_at(self, level: int) -> Stats:
        """Gets stats at a given level between 1 and 15.

        Args:
            level: The level at which to get stats at.

        Example:
            >>> from py_unite_db import UniteDb
            >>> UniteDb().pokemon[0].stats_at(7)
            Stats(hp=3823, attack=293, defense=109, sp_attack=47, sp_defense=76, crit=5, cooldown_reduction=10, lifesteal=0)


        If the level given is less than 1, gets level 1.
        If the level given is more than 15, gets level 15.
        """  # noqa: E501
        level = min(POKEMON_MAX_LEVEL, max(POKEMON_MIN_LEVEL, level))
        return self.stats[level - 1]

    def name_at(self, level: int) -> str:
        """Gets the name of the pokemon at a given level.

        This is only really useful if the Pokemon has evolutions.
        For example, Blastoise only becomes Blastoise at level 9.

        Example:
            >>> from py_unite_db import UniteDb
            >>> UniteDb().pokemon[1].name
            'Blastoise'
            >>> UniteDb().pokemon[1].name_at(1)
            'Squirtle'
            >>> UniteDb().pokemon[1].name_at(5)
            'Wartortle'
            >>> UniteDb().pokemon[1].name_at(9)
            'Blastoise'

        If the level given is less than 1, gets level 1.
        If the level given is more than 15, gets level 15.
        """
        # if no evolutions or level is larger than or equal to final
        if not self.evolution or level >= self.evolution[-1].level:
            return self.name
        # if more than one evolution and middle stage
        elif level >= self.evolution[0].level:
            return self.evolution[-1].name
        # if one evolution or base stage
        return self.evolution[0].name

    @staticmethod
    def _transform(v: dict[str, Any]) -> dict[str, Any]:
        return _flatten(v, "tags")



import pytest
from pydantic.error_wrappers import ValidationError

from py_unite_db import UniteDb

unite_db = UniteDb()


@pytest.mark.webtest
def test_pokemon_change():
    try:
        unite_db.pokemon
    except ValidationError as e:
        pytest.fail(e)


@pytest.mark.webtest
def test_held_items_change():
    try:
        unite_db.held_items
    except ValidationError as e:
        pytest.fail(e)


@pytest.mark.webtest
def test_battle_items_change():
    try:
        unite_db.battle_items
    except ValidationError as e:
        pytest.fail(e)



from pathlib import Path

from py_unite_db import BASE_URL

from .mock_unite_db import MockUniteDb


def test_endpoint_cache():
    mock_unite_db = MockUniteDb()
    aeos_cookie = mock_unite_db.held_items[0]

    assert aeos_cookie.name == "Aeos Cookie"

    # now change the data at "held_items.json"

    mock_unite_db._adapter.register_uri(
        "GET",
        f"{BASE_URL}/held_items.json",
        json=MockUniteDb._load_json(
            Path(__file__).resolve().parent / "responses/held_items_2.json"
        ),
    )

    # the cached version should not change
    assert mock_unite_db.held_items[0].name == "Aeos Cookie"

    # now remove cache (forcing the refetch)
    del mock_unite_db.held_items

    # value is now new value
    assert mock_unite_db.held_items[0].name == "Assault Vest"


def test_pokemon_cache():
    mock_unite_db = MockUniteDb()

    assert mock_unite_db.pokemon[0].name == "Blastoise"
    assert mock_unite_db.pokemon[0].stats[0].hp == 3225

    # now change the data at "pokemon.json" and "stats.json"

    for endpoint in ("pokemon", "stats"):
        mock_unite_db._adapter.register_uri(
            "GET",
            f"{BASE_URL}/{endpoint}.json",
            json=MockUniteDb._load_json(
                Path(__file__).resolve().parent / f"responses/{endpoint}_2.json"
            ),
        )

    # the cached version should not change
    assert mock_unite_db.pokemon[0].name == "Blastoise"
    assert mock_unite_db.pokemon[0].stats[0].hp == 3225

    # now remove cache (forcing the refetch)
    del mock_unite_db.pokemon

    # value is now new value
    assert mock_unite_db.pokemon[0].name == "Absol"
    assert mock_unite_db.pokemon[0].stats[0].hp == 3000



from typing import Any, Callable


def _flatten(d: dict[str, Any], *keys: str) -> dict[str, Any]:
    """Pulls up values from nested dict into outer dict.

    The keys are evaluated in the order they are provided to the function.
    Pulled up values replace those that already exist in the outer dict.
    The key original key provided will be removed unless it is pulled up from the
    nested dict.

    Example:
        >>> flatten({"a": "b", "c": {"c": "d"}}, "c")
        {"a": "b", "c": "d"}
    """
    for key in keys:
        if isinstance(d.get(key), dict):
            d |= d.pop(key)
    return d


def _squash(
    d: dict[str, Any], new_key: str, cond: Callable[[str], bool]
) -> dict[str, Any]:
    """Squash all keys that match the given condition into a list of their values."""
    values = [v for k, v in d.items() if cond(k)]
    return {k: v for k, v in d.items() if not cond(k)} | {new_key: values}


def _percent_as_float(percent: str) -> float:
    """Converts a percentage formatted string to float

    If it is not a percentage, just cast to float.

    Args:
        percent: the percentage formatted string to convert
    """
    return float(percent.rstrip("%")) / 100 if percent.endswith("%") else float(percent)


def _pretty_float(num: float) -> str:
    """Converts a float to a string.

    If the float is equal to its integer representation, don't keep the trailing 0.
    """
    return str(int(num) if int(num) == num else num)


def _index_by(list_: list[dict[str, Any]], key: str) -> dict[str, dict[str, Any]]:
    """Takes list of `dict`s and indexes by a key rather than list index.

    If the key is not in the inner `dict`, it is ignored.
    """
    return {d[key]: d for d in list_ if key in d.keys()}



from .mock_unite_db import MockUniteDb

eject_button = MockUniteDb().battle_items[0]


def test_attributes():
    assert eject_button.name == "Eject Button"
    assert eject_button.display_name == "Eject Button"
    assert eject_button.description == "Quickly move in a designated direction."
    assert eject_button.tier == "S"
    assert eject_button.cooldown == 70
    assert eject_button.unlock_level == 11



from .mock_unite_db import MockUniteDb

mock_unite_db = MockUniteDb()


def test_name():
    assert mock_unite_db.pokemon[0].name == "Blastoise"


def test_name_at():
    assert mock_unite_db.pokemon[0].name_at(0) == "Squirtle"
    assert mock_unite_db.pokemon[0].name_at(1) == "Squirtle"

    assert mock_unite_db.pokemon[0].name_at(5) == "Wartortle"

    assert mock_unite_db.pokemon[0].name_at(9) == "Blastoise"
    assert mock_unite_db.pokemon[0].name_at(16) == "Blastoise"


def test_stats():
    # using indexes
    assert mock_unite_db.pokemon[0].stats[0].hp == 3225
    assert mock_unite_db.pokemon[0].stats[14].hp == 9800

    # using `stats_at` method
    assert mock_unite_db.pokemon[0].stats_at(1).hp == 3225
    assert mock_unite_db.pokemon[0].stats_at(15).hp == 9800
    # test round up/down
    assert mock_unite_db.pokemon[0].stats_at(0).hp == 3225
    assert mock_unite_db.pokemon[0].stats_at(16).hp == 9800


def test_stars():
    assert mock_unite_db.pokemon[0].stars.combat == 4


def test_evolutions():
    assert mock_unite_db.pokemon[0].evolution[0].name == "Squirtle"
    assert mock_unite_db.pokemon[0].evolution[1].name == "Wartortle"



from abc import ABC, abstractmethod
from typing import Any

from ..utils import _percent_as_float, _pretty_float
from .base_model import BaseModel
from .item import Item

HELD_ITEM_MIN_LEVEL = 1
HELD_ITEM_MAX_LEVEL = 30


class ItemEffect(ABC, BaseModel):
    """Class representing the effect a held item gives."""

    label: str

    @abstractmethod
    def description_at(self, level: int) -> str:
        """Calculates the description of the item at a given level.

        Args:
            level: The item level to calculate the description at. (Between 1-30)

        Returns:
            The description at the given level. e.g `Max HP +240`
        """

    @abstractmethod
    def _effect_at(self, level: int) -> float:
        ...

    def effect_at(self, level: int) -> float:
        """Calculates the effect of the item at a given level.

        Args:
            level: The item level the calculate the effect at. (Between 1-30)

        Returns:
            The weight of effect at the given level e.g `240`
        """
        return self._effect_at(
            min(HELD_ITEM_MAX_LEVEL, max(HELD_ITEM_MIN_LEVEL, level))
        )


class LevelItemEffect(ItemEffect):
    level1: str
    level10: str
    level20: str

    def _str_effect_at(self, level: int) -> str:
        if level >= 20:
            return self.level20
        elif level >= 10:
            return self.level10
        else:
            return self.level1

    def _effect_at(self, level: int) -> float:
        return _percent_as_float(self._str_effect_at(level))

    def description_at(self, level: int) -> str:
        return f"{self._str_effect_at(level)} {self.label}"


class ScalingItemEffect(ItemEffect):
    percent: bool
    initial: int
    start: int
    skip: int
    increment: float

    def _effect_at(self, level: int) -> float:
        # ensure the level given is valid
        return self.initial + self.increment * ((level - 1) // (self.skip + 1))

    def description_at(self, level: int) -> str:
        return f"{self.label} +{_pretty_float(self.effect_at(level))}"


class HeldItem(Item):
    """Class representing a held item.

    Get all held items from the api with
    [`UniteDb.held_items`][py_unite_db.UniteDb].

    In the game, you can select up to 3 held items in battle.
    Held items can be upgraded up to level 30.
    """

    effect: LevelItemEffect
    """The effect of the held item"""
    scaling_effects: list[ScalingItemEffect]

    @staticmethod
    def _transform(v: dict[str, Any]) -> dict[str, Any]:
        v["description"] = v.pop("description1")
        v["scaling_effects"] = v.pop("stats")
        # make base for main effect
        v["effect"] = {"label": v.get("description3", "")} | {
            k: v.get(k, "") for k in ("level1", "level10", "level20")
        }
        return v



from typing import Any

from .base_model import BaseModel


class Stats(BaseModel):
    """Stats of a pokemon at a particular level

    View all stats [here](https://unite-db.com/stats) on the unite-db website.
    """

    # absolute
    hp: int
    attack: int
    defense: int
    sp_attack: int
    sp_defense: int
    # percentages
    crit: int
    cooldown_reduction: int
    lifesteal: int

    @staticmethod
    def _transform(v: dict[str, Any]) -> dict[str, Any]:
        v["cooldown_reduction"] = v.pop("cdr")  # make cdr more user friendly
        return v



import json
from pathlib import Path
from typing import Any, Union

import requests
import requests_mock

from py_unite_db import BASE_URL, UniteDbBase


class MockUniteDb(UniteDbBase):
    """
    Mock UniteDb client. Routes all requests through mock adapter to load json from
    snapshots dir.
    """

    _adapter: requests_mock.Adapter

    def __init__(self):
        session = requests.Session()
        self._adapter = requests_mock.Adapter()
        session.mount("https://", self._adapter)

        for name in self._get_endpoints():
            self._register_endpoint(f"{name.lstrip('_')}.json")

        super().__init__(BASE_URL, client=session)

    @staticmethod
    def _load_json(path: Path) -> Union[list[Any], dict[str, Any]]:
        with open(path, "r") as f:
            return json.load(f)

    def _register_endpoint(self, endpoint: str):
        response_path = Path(__file__).resolve().parent / f"responses/{endpoint}"

        self._adapter.register_uri(
            "GET",
            f"{BASE_URL}/{endpoint}",
            json=self._load_json(response_path) if response_path.exists() else {},
        )



from abc import ABC
from functools import wraps
from typing import Any, Callable, Generic, Optional, Type, TypeVar

from uplink import Consumer, get, returns

T = TypeVar("T")


class endpoint(property, Generic[T]):
    """Decorator that automatically interfaces with ApiBase and caches return.

    It allows a restricted, but very clean syntax for defining api endpoints.

    Acts like a `property`. This decorator should only really be used inside
    `ApiBase` as it assumes a `_consumer` attribute and the relevant
    `get_` methods to be bound to the consumer.

    Example:

    ```python
    class Api(ApiBase):
        def __init__(self):
            super().__init__(base_url="https://api.com")

        @endpoint
        def colours(self) -> list[str]:
            \"\"\"Some colours to fetch at https://api.com/colours.json\"\"\"

    >>> Api().colours
    ["red", "yellow", "blue"]
    ```
    """

    _cached: Optional[T] = None

    def __init__(self, f: Callable[[Any], T]):
        @wraps(f)
        def _get(obj: ApiBase):
            """Fetches cached value if exists, else fetch from"""
            self._cached = (
                self._cached
                if self._cached is not None
                else getattr(obj._consumer, f"get_{f.__name__.lstrip('_')}")()
            )
            return self._cached

        def _del(obj: ApiBase):
            """Removes value from cache"""
            self._cached = None

        super().__init__(fget=_get, fdel=_del)


class ApiBase(ABC):
    """Finds all `endpoints` and constructs a `Consumer` which has all of the
    required `get_` methods for the endpoints.
    """

    _consumer: Consumer

    def __init__(self, base_url: str, client=None, consumer: Type[Consumer] = Consumer):
        """
        Args:
            base_url: Base url to be passed to `uplink.Consumer`
            client: Client to be passed to `uplink.Consumer`
            consumer: Comsumer subclass for the consumer to extend.
        """
        # construct meta type with all methods required by endpoints
        # instantiate, and set to `_consumer` attribute.
        self._consumer = type(
            "ConsumerMeta",
            (consumer,),
            {
                # desugared decorators on empty function
                # strip leading `_` so private functions can be used with `endpoint`
                f"get_{name.lstrip('_')}": self.__get_wrapped_method(
                    name,
                    # Make sure we tell uplink what type response we expect
                    # get this from the function type annotation
                    getattr(self.__class__, name).fget.__annotations__.get("return"),
                )
                for name in self._get_endpoints()
            },
        )(base_url, client=client)

    def __get_wrapped_method(self, name: str, rtype: Optional[type]):
        if rtype is None:
            return get(f"{name.lstrip('_')}.json")(lambda _: ...)
        else:
            return returns.json(rtype)(self.__get_wrapped_method(name, None))

    def _get_endpoints(self) -> list[str]:
        """Gets all endpoint decorator names in class and subclasses"""
        return [
            name
            for name in dir(self)
            if type(getattr(self.__class__, name, None)) == endpoint
        ]



import requests

from . import models, types
from .unite_db import UniteDbBase

__version__ = "0.3.0"

BASE_URL = "https://unite-db.com"


class UniteDb(UniteDbBase):
    """Wrapper for the Unite DB rest API.

    Example:
        >>> from py_unite_db import UniteDb
        >>> unite_db = UniteDb()
        >>> unite_db.pokemon[0].name
        'Absol'
        >>> unite_db.held_items[0].name
        'Aeos Cookie'
    """

    def __init__(self):
        super().__init__(BASE_URL, requests.Session())


__all__ = ["BASE_URL", "UniteDb", "types", "models"]



from typing import Literal

Tier = Literal["S", "A", "B", "C", "D", "E", "T", "Y"]
DamageType = Literal["Physical", "Special"]
Range = Literal["Ranged", "Melee"]
Difficulty = Literal["Novice", "Intermediate", "Expert"]
Role = Literal["All-Rounder", "Attacker", "Defender", "Speedster", "Supporter"]
Stat = Literal["Attack", "Sp. Attack", "Defense", "Sp. Defense", "HP", "Speed"]



from functools import cached_property

from .api_base import ApiBase, endpoint
from .models import BattleItem, HeldItem, Pokemon


class UniteDbBase(ApiBase):
    """
    All attributes fetched from the API are cached by default.

    For example, given a instance of this object `unite_db`, you can explicitly delete
    the cache of pokemon by doing:

    ```python
    del unite_db.pokemon
    ```

    And on the next

    ```python
    unite_db.pokemon
    ```

    The value will be pulled freshly from the API.

    It will be very rare you need to this; the API is not updated that frequently.
    """

    def __init__(self, base_url: str, client=None):
        super().__init__(base_url, client)

    @endpoint
    def held_items(self) -> list[HeldItem]:  # type: ignore[empty-body]
        """Get all held items.

        Returns list of [`HeldItem`][py_unite_db.models.HeldItem].
        """

    @endpoint
    def battle_items(self) -> list[BattleItem]:  # type: ignore[empty-body]
        """Get all battle items.

        Returns list of [`BattleItem`][py_unite_db.models.BattleItem].
        """

    @cached_property
    def pokemon(self) -> list[Pokemon]:
        """Get all pokemon.

        Returns list of [`Pokemon`][py_unite_db.models.Pokemon].
        """
        # clear endpoint caches
        # NOTE: Allow separation of cached and uncached endpoints to avoid this?
        del self._pokemon
        del self._stats

        # Combine `_pokemon` and `_stats`
        # get stats by pokemon name
        stats = {stat.get("name", ""): stat["level"] for stat in self._stats.json()}
        pokemon = self._pokemon.json()
        # rename default "stats" to "stars"
        for p in pokemon:
            p["stars"] = p.pop("stats", None)
        # join pokemon and stats together, and parse as `Pokemon` model
        return [Pokemon.parse_obj(p | {"stats": stats.get(p["name"])}) for p in pokemon]

    @endpoint
    def _pokemon(self):  # type: ignore[empty-body]
        """Private pokemon getter.

        Used in background by [py_unite_db.unite_db.UniteDbBase.pokemon]][]
        """

    @endpoint
    def _stats(self):  # type: ignore[empty-body]
        """Private stat getter.

        Used in background by [py_unite_db.unite_db.UniteDbBase.pokemon]][]
        """



from typing import Any

from pydantic import Field

from .item import Item


class BattleItem(Item):
    """Class representing a Battle Item.

    Get all battle items from the api with
    [`UniteDb.battle_items`][py_unite_db.UniteDb].

    In the game, you can select one battle item per battle.
    """

    cooldown: int = Field(..., description="The cooldown of the item in seconds.")
    unlock_level: int = Field(
        ..., description="The trainer level at which this item is unlocked."
    )

    @staticmethod
    def _transform(v: dict[str, Any]) -> dict[str, Any]:
        v["unlock_level"] = v.pop("level")
        return v



from .battle_item import BattleItem
from .held_item import HeldItem, ItemEffect, LevelItemEffect, ScalingItemEffect
from .item import Item
from .pokemon import Evolution, Pokemon, Stars
from .stats import Stats

__all__ = [
    "BattleItem",
    "HeldItem",
    "Item",
    "ItemEffect",
    "ScalingItemEffect",
    "LevelItemEffect",
    "Pokemon",
    "Stars",
    "Evolution",
    "Stats",
]



from typing import Any

from pydantic import BaseModel as PydanticBaseModel
from pydantic import root_validator


class BaseModel(PydanticBaseModel):
    """Globally configured Pydantic base model used by api wrapper."""

    class Config:
        frozen = True  # Disable mutation of attributes and make hashable

    @root_validator(pre=True)
    def _flatten_tags(cls, v: dict[str, Any]) -> dict[str, Any]:
        return cls._transform(v)

    @staticmethod
    def _transform(v: dict[str, Any]) -> dict[str, Any]:
        """Override if need to transform the json response before unpacking

        By default doesn't transform anything.
        """
        return v
